"Collection of various constants."

# Site options
EXPORT_ROOT_PATH_OPTION_KEY = "export_root_path"
EXPORT_ROOT_PATH_OPTION_DEFAULT_VALUE = ""
DEFAULT_LAYOUT_OPTION_KEY = "default_layout"
DEFAULT_LAYOUT_OPTION_DEFAULT_VALUE = None
DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_KEY = "default_page_output_filename"
DEFAULT_PAGE_OUTPUT_FILENAME_OPTION_DEFAULT_VALUE = "index.html"
PAGE_OUTPUT_FILE_EXTENSION_OPTION_KEY = "page_output_file_extension"
PAGE_OUTPUT_FILE_EXTENSION_OPTION_DEFAULT_VALUE = "html"
AUTO_EXPORT_FILES_OPTION_KEY = "auto_export_files"
AUTO_EXPORT_FILES_OPTION_DEFAULT_VALUE = True
